# Implementation on R of the COARE 3.0 bulk flux algorithm and of a salinity-dependent modified
# version of the algorithm.
# Based on the FORTRAN code written by Frank Bradley and obtained from
# ftp://ftp1.esrl.noaa.gov/psd3/cruises/NTAS_2009/RHB/Scientific_analysis/programs/VOCALS2008_programs_leg1/coare/bulkalg/cor3_0/
# See Fairall et al. (2003), J. Climate, 16, 571-591.
#
# Author: Jordi Prats (jordi.prats@irstea.fr)
# Date of creation: 3 November 2017
# Last modification: 3 November 2017
#
#-----------------------------------------------------------------------------------
# Description of the script 
#-----------------------------------------------------------------------------------
# This script contains two types of function:
# 1) Auxiliary functions
# 2) Main functions - actual implementation of the original and modified COARE algorithms
#
# Auxiliary functions
# -------------------
# These are functions used by the main functions to calculate several parameters and physical
# properties. They include the following:
# - dynamic_viscosity: salinity and temperature dependent calculation of dynamic viscosity of
# water
# - gravity: calculation of the gravity constant g as a function of latitude
# - psiu: evaluation of the stability function for wind speed
# - psit: evaluation of the stability function for scalars
# - pv_seawater: calculation of the saturated water vapour pressure as a function of 
# temperature and salinity
# - salinity_exp_coeff: calculation of the salinity expansion coefficient as a function of
# temperature and salinity
# - specific_heat: calculation of the specific heat of water as a function of temperature and
# salinity
# - thermal_exp_coeff: calculation of the thermal expansion coefficient as a function of
# temperature and salinity
# - water_density: calculation of the density of water as a function of temperature and salinity
#
# Main functions
# --------------
# The script contains two implementations of the COARE 3.0 bulk flux algorithm:
# - coare_3_0: Implementation of the original COARE 3.0 bulk flux algorithm (Fairall et al. 
# 2003). Translation of the original FORTRAN code to R.
# - coare_3_0b: Modification of the function coare_3_0 to include temperature and salinity
# dependence in the calculation of the physical properties of water.
#
# Input data
# ----------
# The functions coare_3_0 and coare_3_0b use the following input data:
# - h:        hour (solar time)
# - t:        water temperature (C)
# - t_air:    air temperature (C)
# - p:        air pressure (hPa)
# - sp:       salinity (g/kg)
# - u:        wind speed (m/s)
# - qair:     specific humidity of air (kg/kg)
# - Rs:       solar radiation (W/m2)
# - Rl:       atmospheric longwave radiation (W/m2)
# - rain:     precipitation (mm/h)
# - zmax:     maximum thickness allowed for the warm layer (m)
# - ztemp:    height of measurement of air temperature (m)
# - zu:       height of measurement of wind speed (m)
# - zq:       height of measurement of humidity (m)
# - tw_depth: depth of the bulk surface temperature measurements (m)
# - jwave:    option used for the calculation of the wave state (0: Charnock (Fairall et al.
#             2003); # 1: Oost et al. (2002); 2: Taylor and Yelland(2001))
#
# Output
# ------
# The functions coare_3_0 and coare_3_0b provide an output data frame with the following data:
# - dTc: cool skin temperature difference (C)
# - tkt: cool skin thickness (m)
# - dTw: total warm layer temperature difference (C)
# - DT:  thickness of warm layer (m)
# - Hs:  sensible heat flux (W/m2)
# - Hl:  latent heat flux (W/m2)
# - tau: wind stress (N/m2)
# - rf:  heat flux due to rainfall (W/m2)
# - t0:  skin temperature (C)
#
# References
# ----------
# Beljaars and Holstag 1991. J. Appl. Meteor., 30, 327-341
# Fairall et al. 1996a. J. Geophys. Res., 101(C1), 1295-1308
# Fairall et al. 1996b. J. Geophys. Res., 101(C2), 3747-3764
# Fairall et al. 2003. J. Climate, 16, 571-591
# Imboden and West 1995. Mixing mechanisms in lakes. In: Physics and chemistry of lakes,
# 83-138
# Moritz (1988) Bulletin Geodesique, 62(3) (Geodesist's Handbook), 356
# Oost et al. 2002. Bound.-Layer Meteor., 103, 409-438
# Sharqawy et al. 2010. Desalin. Water Treat., 16, 354-380
# Taylor and Yelland 2001. J. Phys. Oceanogr. 31, 572-590
#
#
# /!\ Warning /!\
# ---------------
# This implementation does not cover the full functionality of the original COARE 3.0 algorithm.
# In particular:
# - The algorithm does not give the option of using skin temperatures to calculate the
# surface fluxes, but only temperature measurements at the depth tw_depth.
# - The study period should start during the night before 6 a.m. (local solar time).
# - The boundary layer depth is fixed to 600 m.
# - The number of output variables is reduced respect to the original implementation.
# /!\ End of warning /!\

#-----------------------------------------------------------------------------------
# Auxiliary functions
#-----------------------------------------------------------------------------------

dynamic_viscosity <- function(t, sp){
  # Calculates the dynamic viscosity of water
  #
  # Based on Sharqawy et al. 2010, Table 5, Eq. (22-23)
  #
  # Args:
  #   t: water temperature (C) in the range 0 < t < 180 C
  #   sp: salinity (g/kg) in the range 0 < sp < 150 g/kg
  #
  # Returns:
  #   m: the dynamic viscosity of water (kg/m.s) 
  
  # Convert units
  S <- sp/1000  # kg/kg
  # Calculate parameters
  A <- 1.541 + 1.998e-2*t - 9.52e-5*t^2
  B <- 7.974 - 7.561e-2*t + 4.724e-4*t^2
  # Calculate pure water dynamic viscosity
  m_0 <- 4.2844e-5 + 1/(0.157*(t + 64.993)^2 - 91.296)
  # Calculate dynamic viscosity with salinity correction
  m <- m_0*(1 + A*S + B*S^2)
  
  return(m)
}

gravity <- function(lat){
  # Calculates g as a function of latitude
  #
  # Based on the 1980 IUGG formula (Moritz, 1988). FORTRAN code by M. Zumberge translated to R.
  # 
  # Args:
  #   lat: latitude (degrees)
  #
  # Returns:
  #   g: gravity constant (m/s2)
  
  # Formula constants
  gamma = 9.7803267715
  c1 = 0.0052790414
  c2 = 0.0000232718
  c3 = 0.0000001262
  c4 = 0.0000000007
  # Convert latitude to radians
  phi = lat * 3.14159265358979/180  # rad
  # Calculate g
  g = gamma * (1.0 + c1 * ((sin(phi))^2) + c2 * ((sin(phi))^4)
               + c3 * ((sin(phi))^6) + c4 * ((sin(phi))^8))
  
  return(g)
}

psiu <- function(zl){
  # Evaluates the stability function for wind speed
  #
  # Args:
  #   zl: stability parameter height/L with L = Obukhov length
  #
  # Returns:
  #   psi: profile stability function
  
  if(zl<0){
    x <- (1 - 15*zl)^0.25  # Kansas unstable
    psik <- 2*log((1+x)/2) + log((1+x^2)/2) - 2*atan(x) + 2*atan(1)
    y <- (1 - 10.15*zl)^0.3333  # Convective (Fairall et al. 1996b)
    psic <- 1.5*log((1 + y + y^2)/3) - sqrt(3)*atan((1 + 2*y)/sqrt(3)) + 4*atan(1)/sqrt(3)
    f <- zl*zl/(1+zl^2)
    psi <- (1 - f)*psik + f*psic
  } else {
    c <- min(50, 0.35*zl)  # Stable (Beljaars and Holstag 1991)
    psi <- -((1 + 1*zl)^1 + 0.6667*(zl - 14.28)/exp(c) + 8.525)
  }
  
  return(psi)
}

psit <- function(zl){
  # Evaluates the stability function for temperature and water vapour
  #
  # Args:
  #   zl: stability parameter height/L with L = Obukhov length
  #
  # Returns:
  #   psi: profile stability function
  
  if(zl<0){
    x <- (1 - 15*zl)^0.5  # Kansas unstable
    psik <- 2*log((1+x)/2)
    y <- (1 - 34.15*zl)^0.3333  # Convective (Fairall et al. 1996b)
    psic <- 1.5*log((1 + y + y^2)/3) - sqrt(3)*atan((1 + 2*y)/sqrt(3)) + 4*atan(1)/sqrt(3)
    f <- zl*zl/(1+zl^2)
    psi <- (1 - f)*psik + f*psic
  } else {
    c <- min(50, 0.35*zl)  # Stable (Beljaars and Holstag 1991)
    psi <- -((1 + 2*zl/3)^1.5 + 0.6667*(zl - 14.28)/exp(c) + 8.525)
  }
  
  return(psi)
}

pv_seawater <- function(pv, t, sp){
  # Calculates the vapour pressure of seawater
  #
  # Based on Sharqawy et al. 2010, Table 7, Eq. (33)
  #
  # Args:
  #   pv: vapour pressure of pure water (hPa)
  #   t: water temperature (C) in the range 0 C < t < 40 C
  #   sp: salinity (g/kg) in the range 0 g/kg < sp < 40 g/kg
  #
  # Returns:
  #   pvs: vapour pressure (hPa)
  
  # Convert units
  pv <- pv*133.322/100  # mm Hg
  # Calculate parameters
  A <- -2.3311e-3 - 1.4799e-4*t - 7.520e-6*t^2 - 5.5185e-8*t^3
  B <- -1.1320e-5 - 8.7086e-6*t + 7.4936e-7*t^2 -2.6327e-8*t^3
  # Calculate seawater vapour pressure
  pvs <- pv + A*sp + B*sp^1.5  # mm Hg
  # Convert units
  pvs <- pvs*100/133.322  # hPa
  
  return(pvs)
}

salinity_exp_coeff <- function(t, sp){
  # Calculates the salinity expansion coefficient
  #
  # Args:
  #   t: water temperature (C)
  #   sp: salinity (g/kg)
  #
  # Returns:
  #   b: salinity expansion coefficient (g/kg)^-1

  # Calculate the extremes of a salinity and density gradient
  sp1 <- sp
  sp2 <- sp + 0.1
  d1 <- water_density(t, sp1)
  d2 <- water_density(t, sp2)
  # Calculate the salinity expansion coefficient (Imboden and West, 1995)
  b <- 1/d1*(d2-d1)/(sp2-sp1)  # (g/kg)^-1
  
  return(b)
}

specific_heat <- function(t, sp){
  # Calculates the specific heat of water
  #
  # Based on Sharqawy et al. 2010, Table 3, Eq. (9)
  #
  # Args:
  #   t: water temperature (C) in the range 0 < t < 180 C
  #   sp: salinity (g/kg) in the range 0 < sp < 180 g/kg
  #
  # Returns:
  #   cp: specific heat of water (J/kg.C)
  
  # Convert units
  t_K <- t + 273.16  # K
  # Calculate parameters
  A <- 5.328 - 9.76e-2*sp + 4.04e-4*sp^2
  B <- -6.913e-3 + 7.351e-4*sp - 3.15e-6*sp^2
  C <- 9.6e-6 - 1.927e-6*sp + 8.23e-9*sp^2
  D <- 2.5e-9 + 1.666e-9*sp - 7.125e-12*sp^2
  # Calculate specific heat
  cp <- (A + B*t_K + C*t_K^2 + D*t_K^3)*1000  # J/kg.C
  
  return(cp)
}

thermal_conductivity <- function(t, sp, p){
  # Calculates the thermal conductivity of water
  #
  # Based on Sharqawy et al. 2010, Table 4, Eq. (14)
  #
  # Args:
  #   t: water temperature (C)
  #   sp: salinity (g/kg)
  #   p: pressure (hPa)
  
  k <- 0.5715*(1+0.003*t-1.025e-5*(t)^2+6.53e-4*p/1e6-0.00029*sp)
  
  return(k)
}

thermal_exp_coeff <- function(t, sp){
  # Calculates the thermal expansion coefficient
  #
  # Args:
  #   t: water temperature (C)
  #   sp: salinity (g/kg)
  #
  # Returns:
  #   a: thermal expansion coefficient (C)^-1
  
  # Calculate the extremes of a temperature and density gradient
  d <- water_density(t, sp)
  t1 <- t + 0.1
  t2 <- t - 0.1
  d1 <- water_density(t1, sp)
  d2 <- water_density(t2, sp)
  # Calculate the thermal expansion coefficient (Imboden and West, 1995)
  a <- -1/d*(d2-d1)/(t2-t1)  # (C)^-1
  
  return(a)
}

water_density <- function(t, sp){
  # Calculates the water density
  #
  # Based on Sharqawy et al. 2010, Table 2, Eq. (6)
  #
  # Args:
  #   t: water temperature (C) in the range -2 < t < 40 C
  #   sp: salinity (g/kg) in the range 0 < sp < 42 g/kg
  #
  # Returns:
  #   dens: water density (kg/m3)
  
  # Calculate parameters
  A <- 0.824493 - 4.0899e-3*t + 7.6438e-5*t^2 - 8.2467e-7*t^3 + 5.3875e-9*t^4
  B <- -5.72466e-3 + 1.0227e-4*t -1.6546e-6*t^2
  C <- 4.8314e-4
  # Calculate pure water density
  dens_0 <- 999.842594 + 6.793952e-2*t - 9.09529e-3*t^2 + 1.001685e-4*t^3 - 1.120083e-6*t^4 + 6.536336e-9*t^5  # kg/m3
  # Calculate seawater density
  dens <- dens_0 + A*sp + B*sp^1.5 + C*sp^2  # kg/m3
  
  return(dens)
}

#-----------------------------------------------------------------------------------
# Main functions
#-----------------------------------------------------------------------------------

coare_3_0 <- function(h, t, t_air, p, sp, u, qair, Rs, Rl, rain, zmax, lat, ztemp, zu, zq, tw_depth, jwave){
  # Calculates cool skin, warm layer and surface fluxes
  #
  # This is a translation to R of the COARE 3.0 algorithm, written in FORTRAN.
  #
  # Args:
  #   h: hour (solar time)
  #   t: water temperature (C)
  #   t_air: air temperature (C)
  #   p: pressure (hPa)
  #   sp: salinity (g/kg)
  #   u: wind speed (m/s)
  #   qair: specific humidity of air (kg/kg)
  #   Rs: solar radiation (W/m2)
  #   Rl: atmospheric longwave radiation (W/m2)
  #   rain: precipitation (mm/h)
  #   zmax: maximum thickness allowed for the warm layer (m)
  #   ztemp: height of measurement of air temperature (m)
  #   zu: height of measurement of wind speed (m)
  #   zq: height of measurement of humidity (m)
  #   tw_depth: depth of the bulk surface temperature measurements (m)
  #   jwave: option used for the calculation of the wave state (0: Charnock (Fairall et al.
  #          2003); 1: Oost et al. (2002); 2: Taylor and Yelland(2001))
  #
  # Returns:
  #   dTc: cool skin temperature difference (C)
  #   tkt: cool skin thickness (m)
  #   dTw: total warm layer temperature difference (C)
  #   DT: thickness of warm layer (m)
  #   Hs: sensible heat flux (W/m2)
  #   Hl: latent heat flux (W/m2)
  #   tau: wind stress (N/m2)
  #   rf: heat flux due to rainfall (W/m2)
  #   t0: skin temperature (C)
  
  # Constants------------------------------------------------------------------
  Ric <- 0.65  # critical Richardson number (Fairall et al. 1996a)
  fdg <- 1
  zi <- 600  # mixed layer height
  von <- 0.4  # von Karman's constant
  beta <- 1.2
  g <- gravity(lat)  # m/s2
  
  b <- rep(7.429e-4, length(t))  # salinity expansion coefficient
  cpa <- rep(1004.67, length(t))  # specific heat of air (J/kg.K)
  cp <- rep(4000, length(t))  # specific heat of water
  dens_w <- rep(1022, length(t))  # water density
  n <- rep(1e-6, length(t))  # kinematic viscosity of water
  k <- rep(0.6, length(t))  # thermal conductivity of water
  
  # Preallocate space for several variables------------------------------------
  # Water properties
  a <- rep(NA, length(t))  # thermal expansion coefficient
  esp_hum_w <- rep(NA, length(t))  # especific humidity at the water/air interphase
  esw <- rep(NA, length(t))  # saturated vapour pressure at the surface
  Le <- rep(NA, length(t))  # latent heat of vaporization
    
  # Heat fluxes
  Hl <- rep(NA, length(t))  # latent heat flux
  Hs <- rep(NA, length(t))  # sensible heat flux
  rf <- rep(NA, length(t))  # rain heat flux
  Rl_up <- rep(NA, length(t))  # upwelling longwave radiation
  Rnl <- rep(NA, length(t))  # net longwave radiation
  
  # Cool skin and warm layer variables
  DT <- rep(NA, length(t))  # warm layer thickness
  dTw <- rep(NA, length(t))  # total warm layer temperature difference
  IS <- rep(NA, length(t))  # heat integral
  It <- rep(NA, length(t))  # momentum integral
  tkt <- rep(NA, length(t))  # cool skin thickness
  dTc <- rep(NA, length(t))  # cool skin temperature difference
  tau <- rep(NA, length(t))  # wind stress
  t0 <- rep(NA, length(t))  # skin temperature
  
  # Calculate several variables----------------------------------------------
  # Air properties
  ta <- t + 273.16  # # air temperature in K
  esa <- (1.0007+3.46e-6*p)*6.1121*exp(17.502*t_air/(240.97+t_air))  # saturated vapour pressure (hPa)
  esp_hum_air_sat <- 0.62197*(esa/(p-0.378*esa))  # saturated air humidity (kg/kg)
  esp_hum_air <- qair  # especific humidity of air (kg/kg)
  dens_air <- p*100/(287.1*ta*(1+0.61*esp_hum_air))  # moist air density (kg/m3)
  na <- 1.326e-5*(1+6.542e-3*t_air+8.301e-6*t_air^2-4.84e-9*t_air^3)  # kinematic viscosity of dry air (m2/s) (Andreas, 1989)
  
  # Wave properties
  # wave height
  hwave <- 0.018*u*u*(1+0.15*u)
  # wave period
  twave <- 0.729*u
  
  # Solar radiation
  # Net shortwave radiation (albedo=0.945)
  Rns <- 0.945*Rs
  
  # Main calculations--------------------------------------------------------
  for (i in 1:length(t)){
    # Warm layer
    # ----------
    if (i==1){
      # Initial surface temperature values
      tsw <- t[i]  # measured water temperature
      dter <- 0.3  # cool skin temperatue difference (C)
      sst <- t[i] - dter  # skin temperature (C) for initial Rnl
    }
    # Longwave irradiance
    Rl_up[i] <- 0.97*5.67e-8*(sst+273.16)^4  # upwelling longwave
    Rnl[i] <- Rl_up[i] - 0.97*Rl[i]  # net longwave
    
    if (i==1){
      # Initialise warm layer values
      DT[i] <- zmax  # m
      IS[i] <- 0
      It[i] <- 0
      dTw[i] <- 0
      start <- 0
      fw <- 0.5
    } else if (h[i]<h[i-1]) {
      # Reset warm layer values at midnight
      IS[i] <- 0
      It[i] <- 0
      dTw[i] <- 0
      start <- 0
      # Initial warm layer parameters expected in early morning
      DT[i] <- zmax  # m
      fw <- 0.5
      tsw <- t[i]
    } else {
      # Warm layer calculations
      dt <- (h[i] - h[i-1])*3600  # time step (s)
      fwRns <- fw*Rns[i]  # total heat absorption in warm layer
      qr_out <- Rnl[i-1] + Hl[i-1] + Hs[i-1] + rf[i-1]  # total cooling at the surface
      if (fwRns - qr_out >= 50 | start!=0){
        # threshold of integration
        start <- 1  # start the daily integration
        It[i] <- It[i-1] + max(0.002, tau[i-1])*dt  # momentum integral
        if(IS[i-1] + (fwRns - qr_out)*dt>0){
          DT[i] <- zmax
          for (j in 1:5){
            # iterate warm layer thickness
            fw <- 1-(0.28*0.014*(1-exp(-DT[i]/0.014))+0.27*0.357*(1-exp(-DT[i]/0.357))+0.45*12.82*(1-exp(-DT[i]/12.82)))/DT[i]  # Soloviev's solar absorption profile
            fwRns <- fw*Rns[i]
            if (IS[i-1] + (fwRns - qr_out)*dt>0){
              DT[i] <- min(zmax, sqrt(2*Ric*cp[i-1])/sqrt(g*a[i-1]*dens_w[i-1])*It[i]/sqrt(IS[i-1] + (fwRns - qr_out)*dt))  # warm layer thickness
            }
          }
        } else {
          # warm layer wiped out
          fw <- 0.75
          DT[i] <- zmax
          fwRns <- fw*Rns[i]
        }
        IS[i] <- IS[i-1] + (fwRns - qr_out)*dt  # heat integral
        if (IS[i]>0){
          dTw[i] <- sqrt(2*a[i-1]*g/(Ric*dens_w[i-1]))/(cp[i-1]^1.5)*IS[i]^1.5/It[i]  # warming
        } else {
          dTw[i] <- 0
        }
      } else {
        DT[i] <- DT[i-1]
        dTw[i] <- dTw[i-1]
        IS[i] <- IS[i-1]
        It[i] <- It[i-1]
      }
    }
    if(DT[i]<tw_depth){
      # sensor deeper than warm layer
      dsea <- dTw[i]
    } else {
      # warming deeper than sensor
      dsea <- dTw[i]*tw_depth/DT[i]  # linear profile assumed
    }
    tsw <- t[i] + dsea  # add warming above sensor to measured temperature
    
    # Atmospheric surface layer
    # -------------------------
    # water properties
    esw[i] <- (1.0007+3.46e-6*p)*6.1121*exp(17.502*tsw/(240.97+tsw))  # hPa
    esw[i] <- esw[i]*0.98  # salinity correction (Kraus 1972, p. 46)
    esp_hum_w[i] <- 0.62197*(esw[i]/(p-0.378*esw[i]))  # kg/kg
    Le[i] <- (2500.9-2.365*tsw)*1000*(1-sp/1000)  # J/kg
    a[i] <- 2.1e-5*(tsw + 3.2)^0.79  # K-1
    wetc <- 0.622*Le[i]*esp_hum_w[i]/(287.1*(tsw+273.16)^2)  # Clausius-Clapeyron
    
    # wave parameters
    cwave <- g*twave[i]/(2*pi)
    lwave <- cwave*twave[i]
    
    # initial guesses
    dter <- 0.3
    dqer <- wetc*dter
    z0 <- 0.0001
    wg <- 0.5
    tkt[i] <- 0.001
    
    # air-sea differences
    du <- sqrt(u[i]^2 + wg^2)  # include gustiness in wind speed difference
    dtemp <- tsw - t_air[i] - 0.0098*ztemp  # potential temperature difference
    dq <- esp_hum_w[i] - esp_hum_air[i]
    
    # neutral coefficients
    u10 <- du*log(10/z0)/log(zu/z0)
    usr <- 0.035*u10
    zo10 <- 0.011*usr*usr/g + 0.11*na[i]/usr
    Cd10 <- (von/log(10/zo10))^2
    Ch10 <- 0.00115
    Ct10 <- Ch10/sqrt(Cd10)
    zot10 <- 10/exp(von/Ct10)
    Cd <- (von/log(zu/zo10))^2
    
    # Grachev and Fairall (JAM, 1997)
    Ct <- von/log(ztemp/zot10)  # temperature transfer coefficient
    CC <- von*Ct/Cd  # z/L vs Rib linear coefficient
    Ribcu <- -zu/(zi*0.004*beta^3)  # Saturation or plateau Rib
    Ribu <- -g*zu*((dtemp-dter)+0.61*ta[i]*dq)/(ta[i]*du^2)
    if (Ribu<0){
      zetu <- CC*Ribu/(1+Ribu/Ribcu)  # unstable G and F
    } else {
      zetu <- CC*Ribu*(1 + 27/9*Ribu/CC)  # Stable
    }
    L10 <- zu/zetu  # MO length
    if (zetu > 50){
      nits <- 1
    } else {
      nits <- 3  # number of iterations
    }
    
    # First guess M-O stability dependent params. (u*, t*, q*) to estimate z0 and z/L
    usr <- du*von/(log(zu/zo10)-psiu(zu/L10))
    tsr <- -(dtemp-dter)*von/(log(ztemp/zot10)-psit(ztemp/L10))
    qsr <- -(dq-dqer)*von/(log(zq/zot10)-psit(zq/L10))
    
    charn <- 0.011
    # modify Charnock parameter for high wind speeds
    if(du>10) {
      charn <- 0.011 + (0.018-0.011)*(du-10)/(18-10)
    }
    if(du>18) {
      charn <- 0.018
    }
    
    for (j in 1:nits){
      # iterate to adjust values of u*, t*, q*, zo, zot, zoq and z/L including cool skin
      if(jwave==0){
        z0 <- charn*usr*usr/g + 0.11*na[i]/usr  # after Smith (1988)
      } else if (jwave==1){
        z0=(50/(2*pi))*lwave*(usr/cwave)^4.5 + 0.11*na[i]/usr # Oost et al.
      } else if(jwave==2) {
        z0=1200.*hwave[i]*(hwave[i]/lwave)**4.5 + 0.11*na[i]/usr # Taylor and Yelland 
      }
      rr <- z0*usr/na[i]
      
      # zoq and zot fitted to results from several ETL cruises
      zoq <- min(1.15e-4, 5.5e-5/rr^0.6)
      zot <- zoq
      
      zl <- von*g*zu*(tsr*(1+0.61*qair[i])+0.61*ta[i]*qsr)/(ta[i]*usr*usr*(1+0.61*qair[i]))
      L <- zu/zl
      psu <- psiu(zu/L)
      pst <- psit(ztemp/L)
      dqer <- wetc*dter
      usr <- du*von/(log(zu/z0)-psiu(zu/L))
      tsr <- -(dtemp-dter)*von/(log(ztemp/zot)-psit(ztemp/L))
      qsr <- -(dq-dqer)*von/(log(zq/zoq)-psit(zq/L))
      bf <- -g/ta[i]*usr*(tsr+0.61*ta[i]*qsr)
      if(bf>0){
        wg <- beta*(bf*zi)^0.333
      } else {
        wg <- 0.2
      }
      du <- sqrt(u[i]^2 + wg^2)  # include gustiness in wind speed
      
      Rl_up[i] <- 0.97*5.67e-8*(tsw-dter+273.16)^4  # Recalculate upwelling longwave
      
      # Cool skin
      # ---------
      Rnl[i] <- Rl_up[i] - 0.97*Rl[i]
      Hsb <- -dens_air[i]*cpa[i]*usr*tsr
      Hlb <- -dens_air[i]*Le[i]*usr*qsr
      qout <- Rnl[i] + Hsb + Hlb
      dels <- Rns[i]*(0.065+11*tkt[i]-6.6e-5/tkt[i]*(1-exp(-tkt[i]/8e-4)))  # Ohlmann, Fairall et al. (1996a), Eq. 16
      qcol <- qout-dels
      bigc <- 16*g*cp[i]*(dens_w[i]*n[i])^3/(k[i]^2*dens_air[i]^2)
      alq <- a[i]*qcol + sp*b[i]*Hlb*cp[i]/Le[i]  # Fairall et al. (1996a), Eq. 8
      if(qcol>0){
        if(alq>=0){
          l <- 6/(1+(bigc*alq/usr^4)^0.75)^0.333  # Fairall et al. (1996a), Eq. 14
          tkt[i] <- l*n[i]/(sqrt(dens_air[i]/dens_w[i])*usr)  # Fairall et al. (1996a), Eq. 12
        } else {
          l <- 6  # to prevent excessive warm skins
          tkt[i] <- min(0.01, l*n[i]/(sqrt(dens_air[i]/dens_w[i])*usr))  # limit tkt
        }
        dter <- qcol*tkt[i]/k[i]  # Fairall et al. (1996a), Eq. 13
      } else {
        dter <- 0
      }
      dqer <- wetc*dter
    }
    
    # surface fluxes and others
    sst <- tsw - dter  # final skin temperature for the current time step
    tau[i] <- dens_air[i]*usr*usr*u[i]/du
    Hs[i] <- -dens_air[i]*cpa[i]*usr*tsr
    Hl[i] <- -dens_air[i]*Le[i]*usr*qsr
    dTc[i] <- dter
    t0[i] <- sst
    
    # rainfall heat flux
    dwat <- 2.11e-5*(ta[i]/273.16)^1.94  # water vapour diffusivity
    dtmp <- (1+3.309e-3*t_air[i]-1.44e-6*t_air[i]^2)*0.02411/(dens_air[i]*cpa[i])  # heat diffusivity
    dqs_dt <- esp_hum_air_sat[i]*Le[i]/(287.1*ta[i]^2)  # Clausius-Calpeyron
    alfac <- 1/(1+(wetc*Le[i]*dwat)/(cpa[i]*dtmp))  # wet bulb factor
    rf[i] <- rain[i]*alfac*cp[i]*((sst-t_air[i])+(esp_hum_w[i]-esp_hum_air[i]-dqer)*Le[i]/cpa[i])/3600
  }
  return(data.frame(dTc, tkt, dTw, DT, Hs, Hl, tau, rf, t0))
}

coare_3_0b <- function(h, t, t_air, p, sp, u, qair, Rs, Rl, rain, zmax, lat, ztemp, zu, zq, tw_depth, jwave){
  # Calculates cool skin, warm layer and surface fluxes (salinity dependent)
  #
  # This is a modification of the original COARE 3.0 algorithm (Fairall et al. 2003) to allow
  # the calculation of the properties of water as a function of salinity.
  #
  # Args:
  #   h: hour (solar time)
  #   t: water temperature (C)
  #   t_air: air temperature (C)
  #   p: pressure (hPa)
  #   sp: salinity (g/kg)
  #   u: wind speed (m/s)
  #   qair: specific humidity of air (kg/kg)
  #   Rs: solar radiation (W/m2)
  #   Rl: atmospheric longwave radiation (W/m2)
  #   rain: precipitation (mm/h)
  #   zmax: maximum thickness allowed for the warm layer (m)
  #   ztemp: height of measurement of air temperature (m)
  #   zu: height of measurement of wind speed (m)
  #   zq: height of measurement of humidity (m)
  #   tw_depth: depth of the bulk surface temperature measurements (m)
  #   jwave: option used for the calculation of the wave state (0: Charnock (Fairall et al.
  #          2003); 1: Oost et al. (2002); 2: Taylor and Yelland(2001))
  #
  # Returns:
  #   dTc: cool skin temperature difference (C)
  #   tkt: cool skin thickness (m)
  #   dTw: total warm layer temperature difference (C)
  #   DT: thickness of warm layer (m)
  #   Hs: sensible heat flux (W/m2)
  #   Hl: latent heat flux (W/m2)
  #   tau: wind stress (N/m2)
  #   rf: heat flux due to rainfall (W/m2)
  #   t0: skin temperature (C)
  
  # Constants------------------------------------------------------------------
  Ric <- 0.65  # critical Richardson number (Fairall et al. 1996a)
  fdg <- 1
  zi <- 600  # mixed layer height
  von <- 0.4  # von Karman's constant
  beta <- 1.2
  g <- gravity(lat)  # m/s2
  
  # Preallocate space for several variables------------------------------------
  # Water properties
  a <- rep(NA, length(t))  # thermal expansion coefficient
  b <- rep(NA, length(t))  # salinity expansion coefficient
  cp <- rep(NA, length(t))  # specific heat of water
  dens_w <- rep(NA, length(t))  # water density
  esp_hum_w <- rep(NA, length(t))  # especific humidity at the water/air interphase
  esw <- rep(NA, length(t))  # saturated vapour pressure at the surface
  k <- rep(NA, length(t))  # thermal conductivity of water
  Le <- rep(NA, length(t))  # latent heat of vaporization
  n <- rep(NA, length(t))  # kinematic viscosity of water
  
  # Heat fluxes
  Hl <- rep(NA, length(t))  # latent heat flux
  Hs <- rep(NA, length(t))  # sensible heat flux
  rf <- rep(NA, length(t))  # rain heat flux
  Rl_up <- rep(NA, length(t))  # upwelling longwave radiation
  Rnl <- rep(NA, length(t))  # net longwave radiation
  
  # Cool skin and warm layer variables
  DT <- rep(NA, length(t))  # warm layer thickness
  dTw <- rep(NA, length(t))  # total warm layer temperature difference
  IS <- rep(NA, length(t))  # heat integral
  It <- rep(NA, length(t))  # momentum integral
  tkt <- rep(NA, length(t))  # cool skin thickness
  dTc <- rep(NA, length(t))  # cool skin temperature difference
  tau <- rep(NA, length(t))  # wind stress
  t0 <- rep(NA, length(t))  # skin temperature

  # Calculate several variables----------------------------------------------
  # Air properties
  ta <- t + 273.16  # # air temperature in K
  esa <- (1.0007+3.46e-6*p)*6.1121*exp(17.502*t_air/(240.97+t_air))  # saturated vapour pressure (hPa)
  esp_hum_air_sat <- 0.62197*(esa/(p-0.378*esa))  # saturated air humidity (kg/kg)
  esp_hum_air <- qair  # especific humidity of air (kg/kg)
  dens_air <- p*100/(287.1*ta*(1+0.61*esp_hum_air))  # air density (kg/m3)
  na <- 1.326e-5*(1+6.542e-3*t_air+8.301e-6*t_air^2-4.84e-9*t_air^3)  # kinematic viscosity of dry air (m2/s) (Andreas, 1989)
  cpa <- 1002.5+275e-6*(t_air+273.16-200)^2  # specific heat of air (J/kg.K)
  
  # Wave properties
  # wave height
  hwave <- 0.018*u*u*(1+0.15*u)
  # wave period
  twave <- 0.729*u
  
  # Solar radiation
  # Net shortwave radiation (albedo=0.945)
  Rns <- 0.945*Rs
  
  # Main calculations--------------------------------------------------------
  for (i in 1:length(t)){
    # Warm layer
    # ----------
    if (i==1){
      # Initial surface temperature values
      tsw <- t[i]  # measured water temperature
      dter <- 0.3  # cool skin temperatue difference (C)
      sst <- t[i] - dter  # skin temperature (C) for initial Rnl
    }
    # Longwave irradiance
    Rl_up[i] <- 0.97*5.67e-8*(sst+273.16)^4  # upwelling longwave
    Rnl[i] <- Rl_up[i] - 0.97*Rl[i]  # net longwave
    
    if (i==1){
      # Initialise warm layer values
      DT[i] <- zmax  # m
      IS[i] <- 0
      It[i] <- 0
      dTw[i] <- 0
      start <- 0
      fw <- 0.5
    } else if (h[i]<h[i-1]) {
      # Reset warm layer values at midnight
      IS[i] <- 0
      It[i] <- 0
      dTw[i] <- 0
      start <- 0
      # Initial warm layer parameters expected in early morning
      DT[i] <- zmax  # m
      fw <- 0.5
      tsw <- t[i]
    } else {
      # Warm layer calculations
      dt <- (h[i] - h[i-1])*3600  # time step (s)
      fwRns <- fw*Rns[i]  # total heat absorption in warm layer
      qr_out <- Rnl[i-1] + Hl[i-1] + Hs[i-1] + rf[i-1]  # total cooling at the surface
      if (fwRns - qr_out >= 50 | start!=0){
        # threshold of integration
        start <- 1  # start the daily integration
        It[i] <- It[i-1] + max(0.002, tau[i-1])*dt  # momentum integral
        if(IS[i-1] + (fwRns - qr_out)*dt>0){
          DT[i] <- zmax
          for (j in 1:5){
            # iterate warm layer thickness
            fw <- 1-(0.28*0.014*(1-exp(-DT[i]/0.014))+0.27*0.357*(1-exp(-DT[i]/0.357))+0.45*12.82*(1-exp(-DT[i]/12.82)))/DT[i]  # Soloviev's solar absorption profile
            fwRns <- fw*Rns[i]
            if (IS[i-1] + (fwRns - qr_out)*dt>0){
              DT[i] <- min(zmax, sqrt(2*Ric*cp[i-1])/sqrt(g*a[i-1]*dens_w[i-1])*It[i]/sqrt(IS[i-1] + (fwRns - qr_out)*dt))  # warm layer thickness
            }
          }
        } else {
          # warm layer wiped out
          fw <- 0.75
          DT[i] <- zmax
          fwRns <- fw*Rns[i]
        }
        IS[i] <- IS[i-1] + (fwRns - qr_out)*dt  # heat integral
        if (IS[i]>0){
          dTw[i] <- sqrt(2*a[i-1]*g/(Ric*dens_w[i-1]))/(cp[i-1]^1.5)*IS[i]^1.5/It[i]  # warming
        } else {
          dTw[i] <- 0
        }
      } else {
        DT[i] <- DT[i-1]
        dTw[i] <- dTw[i-1]
        IS[i] <- IS[i-1]
        It[i] <- It[i-1]
      }
    }
    if(DT[i]<tw_depth){
      # sensor deeper than warm layer
      dsea <- dTw[i]
    } else {
      # warming deeper than sensor
      dsea <- dTw[i]*tw_depth/DT[i]  # linear profile assumed
    }
    tsw <- t[i] + dsea  # add warming above sensor to measured temperature
    
    # Atmospheric surface layer
    # -------------------------
    # water properties
    esw[i] <- (1.0007+3.46e-6*p)*6.1121*exp(17.502*tsw/(240.97+tsw))
    esw[i] <- pv_seawater(esw[i], tsw, sp)  # salinity correction
    esp_hum_w[i] <- 0.62197*(esw[i]/(p-0.378*esw[i]))  # kg/kg
    Le[i] <- (2500.9-2.365*tsw)*1000*(1-sp/1000)  # J/kg
    a[i] <- thermal_exp_coeff(tsw, sp)  # K-1
    b[i] <- salinity_exp_coeff(tsw, sp)  # (g/kg)^-1
    k[i] <- thermal_conductivity(tsw, sp, p)  # W/m.C
    dens_w[i] <- water_density(tsw, sp)  # kg/m3
    m <- dynamic_viscosity(tsw, sp)  # kg/m.s
    n[i] <- m/dens_w[i]  # kinematic viscosity (m2/s)
    cp[i] <- specific_heat(tsw, sp)  # J/Kg.K
    wetc <- 0.622*Le[i]*esp_hum_w[i]/(287.1*(tsw+273.16)^2)  # Clausius-Clapeyron
    
    # wave parameters
    cwave <- g*twave[i]/(2*pi)
    lwave <- cwave*twave[i]
    
    # initial guesses
    dter <- 0.3
    dqer <- wetc*dter
    z0 <- 0.0001
    wg <- 0.5
    tkt[i] <- 0.001
    
    # air-sea differences
    du <- sqrt(u[i]^2 + wg^2)  # include gustiness in wind speed difference
    dtemp <- tsw - t_air[i] - 0.0098*ztemp  # potential temperature difference
    dq <- esp_hum_w[i] - esp_hum_air[i]
    
    # neutral coefficients
    u10 <- du*log(10/z0)/log(zu/z0)
    usr <- 0.035*u10
    zo10 <- 0.011*usr*usr/g + 0.11*na[i]/usr
    Cd10 <- (von/log(10/zo10))^2
    Ch10 <- 0.00115
    Ct10 <- Ch10/sqrt(Cd10)
    zot10 <- 10/exp(von/Ct10)
    Cd <- (von/log(zu/zo10))^2
    
    # Grachev and Fairall (JAM, 1997)
    Ct <- von/log(ztemp/zot10)  # temperature transfer coefficient
    CC <- von*Ct/Cd  # z/L vs Rib linear coefficient
    Ribcu <- -zu/(zi*0.004*beta^3)  # Saturation or plateau Rib
    Ribu <- -g*zu*((dtemp-dter)+0.61*ta[i]*dq)/(ta[i]*du^2)
    if (Ribu<0){
      zetu <- CC*Ribu/(1+Ribu/Ribcu)  # unstable G and F
    } else {
      zetu <- CC*Ribu*(1 + 27/9*Ribu/CC)  # Stable
    }
    L10 <- zu/zetu  # MO length
    if (zetu > 50){
      nits <- 1
    } else {
      nits <- 3  # number of iterations
    }
    
    # First guess M-O stability dependent params. (u*, t*, q*) to estimate z0 and z/L
    usr <- du*von/(log(zu/zo10)-psiu(zu/L10))
    tsr <- -(dtemp-dter)*von/(log(ztemp/zot10)-psit(ztemp/L10))
    qsr <- -(dq-dqer)*von/(log(zq/zot10)-psit(zq/L10))
    
    charn <- 0.011
    # modify Charnock parameter for high wind speeds
    if(du>10) {
      charn <- 0.011 + (0.018-0.011)*(du-10)/(18-10)
    }
    if(du>18) {
      charn <- 0.018
    }
    
    for (j in 1:nits){
      # iterate to adjust values of u*, t*, q*, zo, zot, zoq and z/L including cool skin
      if(jwave==0){
        z0 <- charn*usr*usr/g + 0.11*na[i]/usr  # after Smith (1988)
      } else if (jwave==1){
        z0=(50/(2*pi))*lwave*(usr/cwave)^4.5 + 0.11*na[i]/usr # Oost et al.
      } else if(jwave==2) {
        z0=1200.*hwave[i]*(hwave[i]/lwave)**4.5 + 0.11*na[i]/usr # Taylor and Yelland 
      }
      rr <- z0*usr/na[i]
      
      # zoq and zot fitted to results from several ETL cruises
      zoq <- min(1.15e-4, 5.5e-5/rr^0.6)
      zot <- zoq
      
      zl <- von*g*zu*(tsr*(1+0.61*qair[i])+0.61*ta[i]*qsr)/(ta[i]*usr*usr*(1+0.61*qair[i]))
      L <- zu/zl
      psu <- psiu(zu/L)
      pst <- psit(ztemp/L)
      dqer <- wetc*dter
      usr <- du*von/(log(zu/z0)-psiu(zu/L))
      tsr <- -(dtemp-dter)*von/(log(ztemp/zot)-psit(ztemp/L))
      qsr <- -(dq-dqer)*von/(log(zq/zoq)-psit(zq/L))
      bf <- -g/ta[i]*usr*(tsr+0.61*ta[i]*qsr)
      if(bf>0){
        wg <- beta*(bf*zi)^0.333
      } else {
        wg <- 0.2
      }
      du <- sqrt(u[i]^2 + wg^2)  # include gustiness in wind speed
      
      Rl_up[i] <- 0.97*5.67e-8*(tsw-dter+273.16)^4  # Recalculate upwelling longwave
      
      # Cool skin
      # ---------
      Rnl[i] <- Rl_up[i] - 0.97*Rl[i]
      Hsb <- -dens_air[i]*cpa[i]*usr*tsr
      Hlb <- -dens_air[i]*Le[i]*usr*qsr
      qout <- Rnl[i] + Hsb + Hlb
      dels <- Rns[i]*(0.065+11*tkt[i]-6.6e-5/tkt[i]*(1-exp(-tkt[i]/8e-4)))  # Ohlmann, Fairall et al. (1996a), Eq. 16
      qcol <- qout-dels
      bigc <- 16*g*cp[i]*(dens_w[i]*n[i])^3/(k[i]^2*dens_air[i]^2)
      alq <- a[i]*qcol + sp*b[i]*Hlb*cp[i]/Le[i]  # Fairall et al. (1996a), Eq. 8
      if(qcol>0){
        if(alq>=0){
          l <- 6/(1+(bigc*alq/usr^4)^0.75)^0.333  # Fairall et al. (1996a), Eq. 14
          tkt[i] <- l*n[i]/(sqrt(dens_air[i]/dens_w[i])*usr)  # Fairall et al. (1996a), Eq. 12
        } else {
          l <- 6  # to prevent excessive warm skins
          tkt[i] <- min(0.01, l*n[i]/(sqrt(dens_air[i]/dens_w[i])*usr))  # limit tkt
        }
        dter <- qcol*tkt[i]/k[i]  # Fairall et al. (1996a), Eq. 13
      } else {
        dter <- 0
      }
      dqer <- wetc*dter
    }
    
    # surface fluxes and others
    sst <- tsw - dter  # final skin temperature for the current time step
    tau[i] <- dens_air[i]*usr*usr*u[i]/du
    Hs[i] <- -dens_air[i]*cpa[i]*usr*tsr
    Hl[i] <- -dens_air[i]*Le[i]*usr*qsr
    dTc[i] <- dter
    t0[i] <- sst

    # rainfall heat flux
    dwat <- 2.11e-5*(ta[i]/273.16)^1.94  # water vapour diffusivity
    dtmp <- (1+3.309e-3*t_air[i]-1.44e-6*t_air[i]^2)*0.02411/(dens_air[i]*cpa[i])  # heat diffusivity
    dqs_dt <- esp_hum_air_sat[i]*Le[i]/(287.1*ta[i]^2)  # Clausius-Calpeyron
    alfac <- 1/(1+(wetc*Le[i]*dwat)/(cpa[i]*dtmp))  # wet bulb factor
    rf[i] <- rain[i]*alfac*cp[i]*((sst-t_air[i])+(esp_hum_w[i]-esp_hum_air[i]-dqer)*Le[i]/cpa[i])/3600
  }
  return(data.frame(dTc, tkt, dTw, DT, Hs, Hl, tau, rf, t0))
}
